/* ne2-mca.c: A ne/2 ethernet driver for linux. */
/*
    Based on the NE200 driver written by Donald Becker (1992-94).
    modified by Wim Dumon (Apr 1996)
    some integration hacks by Chris Beauregard (Nov 1996)

    This software may be used and distributed according to the terms
    of the GNU Public License, incorporated herein by reference.

    The author may be reached as wimpie@linux.cc.kuleuven.ac.be
    

    Currently suported : NE/2
    This patch was never tested on other MCA-ethernet adaptors, but it
    might work. Just give it a try and let me know if you have problems.
    Also mail me if it realy works, please !
    

*/

/*  WARNING
    -------
    This is alpha-test software.  It is not guaranteed to work. As a
    matter of fact, I'm quite sure there are *LOTS* of bugs in here. I
    would like to hear from you if you use this driver, even if it works.
    If it doesn't work, be sure to send me a mail with the problems !
*/


static const char *version =
    "ne2-mca.c:v0.01 Apr 29 1996 Wim Dumon (Wim.Dumon@linux.cc.kuleuven.ac.be)\n";


#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <asm/system.h>
#include <asm/io.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include "8390.h"
#include <linux/mca.h>

/* Some defines that people can play with if so inclined. */

/* Do we perform extra sanity checks on stuff ? */
#undef NE_SANITY_CHECK

/* Do we implement the read before write bugfix ? */
/* #define NE_RW_BUGFIX */

/* Do we have a non std. amount of memory? (in units of 256 byte pages) */
/* #define PACKETBUF_MEMSIZE	0x40 */ /*ne2 : What should we put here ?*/

/* ---- No user-serviceable parts below ---- */


#define NE_BASE	 (dev->base_addr)
#define NE_CMD	 	0x00
#define NE_DATAPORT	0x10	/* NatSemi-defined port window offset. */
/*this surprised me : the reset port in the NE/2 is different from the one
  in the NE2000.  This is probably the biggest difference between the 2
  drivers*/
/*NE2000:#define NE_RESET	0x1f*/	/* Issue a read to reset, a write to clear. */
#define NE_RESET	0x20	/* Issue a read to reset, a write to clear. */
/*The address regions specified by the .ADF-file are 0X30 in size. This is
  another strange diff between the NE/2 and NE2000. Why do we have more
  IO-ports here ? Unused functionality ? ;-)*/
/*#define NE_IO_EXTENT	0x20*/
#define NE_IO_EXTENT	0x30

#define NE1SM_START_PG	0x20	/* First page of TX buffer */
#define NE1SM_STOP_PG 	0x40	/* Last page +1 of RX ring */
#define NESM_START_PG	0x40	/* First page of TX buffer */
#define NESM_STOP_PG	0x80	/* Last page +1 of RX ring */

#define DCR        0x0e   /*Copied from the crynwr driver*/
#define RBCR0      0x0a   /*I don't think I used them in this piece of src*/
#define RBCR1      0x0b
#define RCRWD      0x0C
#define TCR        0x0d
#define START_PG   0x46
#define PSTART     0x01
#define BNRY       0x03
#define STOP_PG    0x80
#define PSTOP      0x02
#define ISR        0x07
#define IMR        0x0f
#define CMDR       0x00
#define CURR       0x07
#define RSAR1      0x09
#define RSAR0      0x08
#define MSK_DMA_RD 0x0a
#define RACK       0x10

int ne2mca_probe(struct device *dev);

static int ne_open(struct device *dev);
static int ne_close(struct device *dev);

static void ne_reset_8390(struct device *dev);
static void ne_get_8390_hdr(struct device *dev, struct e8390_pkt_hdr *hdr,
			  int ring_page);
static void ne_block_input(struct device *dev, int count,
			  struct sk_buff *skb, int ring_offset);
static void ne_block_output(struct device *dev, const int count,
		const unsigned char *buf, const int start_page);



int ne2mca_probe(struct device *dev)
{
  int slot=0, base_addr, irq, i;
  unsigned char pos0;
  int start_page=0x40, stop_page=0x80; /*Just took the same values as for
                                         the NE2000. correct ??*/
  unsigned char SA_prom[32];
  char tmp_POS;
  const char *name = "NE2";

  /*ei_debug=7;*/

  /*From the .ADF file :*/
  unsigned int addresses[7]=  
    {0x1000, 0x2020, 0x8020, 0xa0a0, 0xb0b0, 0xc0c0, 0xc3d0};
  int irqs[4]={3, 4, 5, 9};

  slot = mca_find_adapter(0x7154, 0); /*Only search for 1 card in this
                                               version */
  if( slot == MCA_NOTFOUND ) {
    return ENODEV;
  }

  printk("%s",version);

  printk("ne2-mca.c: Found ne/2 card in slot %d ; ", slot + 1);
  pos0 = mca_read_stored_pos( slot, 2 );
  printk( "POS: %X\n", pos0 );

  /*We found an NE/2. Lets grab the irq line and base IO ...*/
  if( !(pos0 % 2) ) {
    printk("ne2-mca.c: NE/2 is disabled.  Ignoring this card ...\n");
    return ENODEV;
  }
  tmp_POS = (pos0 & 0xe)>>1;
  base_addr=addresses[tmp_POS - 1];
  tmp_POS = (pos0 & 0x60)>>5;
  irq=irqs[(int)tmp_POS];

  /*Reset the card the way they do it in the Crynwr packet driver*/
  for (i=0; i<8; i++) outb(0x0,base_addr + NE_RESET); 
  inb(base_addr + NE_RESET);
  outb(0x21, base_addr + NE_CMD);
  if (inb(base_addr + NE_CMD)!=0x21) {
    printk("NE2 not responding\n");  /*Hell, can this ever happen ?*/
    return ENODEV;
  }


#ifdef disabled_code
  /*In the crynwr sources they do a RAM-test here. I skip it. I suppose my
    RAM is okay.  Maybe your memory is broken.  Then this test will fail.
    So you won't be able to use your card.  But if I do not test, you won't
    be able to use your card, neither.  So this test is nonsense.
  */
  /*initializing for RAM-test (loopback mode etc)*/
  { struct {unsigned char value, offset; } program_seq[] = {
      {0x49,      DCR},      /* Set word-wide (0x49) access. */
      {0x00,      RBCR0},    /* Clear the count regs. */
      {0x00,      RBCR1},
      {0x04,      RCRWD},
      {0x02,      TCR},
      {START_PG,  PSTART},
      {START_PG,  BNRY},
      {STOP_PG,   PSTOP},
      {0xFF,      ISR},
      {0x1F,      IMR},
      {0x61,      CMDR},
      {START_PG + 1,      CURR},
      {0x21,      CMDR},
    };
    for (i = 0; i < sizeof(program_seq)/sizeof(program_seq[0]); i++)
        outb_p(program_seq[i].value, base_addr + program_seq[i].offset);
  }
  /*RAM-test comes here ... but not (yet) implemented !*/
  outb_p(0x21, base_addr + CMDR);

  { struct {unsigned char value, offset; } program_seq[] = {
      {0x12,      RBCR0},
      {0x00,      RBCR1},
      {0x00,      RSAR0},
      {0x00,      RSAR1},
      {MSK_DMA_RD,CMDR},
    };
    for (i = 0; i < sizeof(program_seq)/sizeof(program_seq[0]); i++)
        outb_p(program_seq[i].value, base_addr + program_seq[i].offset);
  }
#endif
    {
        struct {unsigned char value, offset; } program_seq[] = {
            {E8390_NODMA+E8390_PAGE0+E8390_STOP, E8390_CMD}, /* Select page 0*/
            {0x49,      EN0_DCFG},      /* Set word-wide (0x49) access. */
            {0x00,      EN0_RCNTLO},    /* Clear the count regs. */
            {0x00,      EN0_RCNTHI},
            {0x00,      EN0_IMR},       /* Mask completion irq. */
            {0xFF,      EN0_ISR},
            {E8390_RXOFF, EN0_RXCR},    /* 0x20  Set to monitor */
            {E8390_TXOFF, EN0_TXCR},    /* 0x02  and loopback mode. */
 /*Whoops?*/{32,        EN0_RCNTLO},
            {0x00,      EN0_RCNTHI},
            {0x00,      EN0_RSARLO},    /* DMA starting at 0x0000. */
            {0x00,      EN0_RSARHI},
            {E8390_RREAD+E8390_START, E8390_CMD},
        };
        for (i = 0; i < sizeof(program_seq)/sizeof(program_seq[0]); i++)
            outb_p(program_seq[i].value, base_addr + program_seq[i].offset);

    }


  /*now read the hardware ethernet address ...*/
  for(i=0; i<6; i++) {
    SA_prom[i]=inb(base_addr + RACK);
    /*printk("%x ", SA_prom[i]);*/
  }  /* This stuff looks like working ...*/

  /*set_address: this is implemented in NS8390_init*/
  /*From now on, there are almost no differences with the NE2000 driver.
    Most important stuff is that we do not need and IRQ-probing*/

  /* We should have a "dev" from Space.c or the static module table. */
  if (dev == NULL) {
    printk("ne2-mca.c: Passed a NULL device.\n");
    dev = init_etherdev(0, 0);
  }
  dev->irq=irq;
  
  /* Snarf the interrupt now.  There's no point in waiting since we cannot
     share and the board will usually be enabled. */
  {
    int irqval = request_irq(dev->irq, ei_interrupt, 0, name, dev);
    if (irqval) {
      printk (" unable to get IRQ %d (irqval=%d).\n", dev->irq, irqval);
      return EAGAIN;
    }
  }

  dev->base_addr = base_addr;

  /* Allocate dev->priv and fill in 8390 specific dev fields. */
  if (ethdev_init(dev)) {
      printk (" unable to get memory for dev->priv.\n");
      free_irq(dev->irq, dev);
      return -ENOMEM;
  }

  request_region(base_addr, NE_IO_EXTENT, name);

  for(i = 0; i < ETHER_ADDR_LEN; i++) {
    printk(" %2.2x", SA_prom[i]);
    dev->dev_addr[i] = SA_prom[i];
  }


  printk("NE/2 ethercard probe at %#4x:", base_addr);
  printk("\n%s: %s found at %#x, using IRQ %d.\n",
         dev->name, name, base_addr, dev->irq);

  ei_status.name = name;
  ei_status.tx_start_page = start_page;
  ei_status.stop_page = stop_page;
  ei_status.word16 = (1==1); /*hehheh, NE2 _is_ 16 bit card*/

  ei_status.rx_start_page = start_page + TX_PAGES; /*I hope this is correct:)*/
#ifdef PACKETBUF_MEMSIZE
    /* Allow the packet buffer size to be overridden by know-it-alls. */
    ei_status.stop_page = ei_status.tx_start_page + PACKETBUF_MEMSIZE;
#endif

  ei_status.reset_8390 = &ne_reset_8390;
  ei_status.block_input = &ne_block_input;
  ei_status.block_output = &ne_block_output;
  ei_status.get_8390_hdr = &ne_get_8390_hdr;
  dev->open = &ne_open;
  dev->stop = &ne_close;
  NS8390_init(dev, 0);
  return 0;


  
}


static int
ne_open(struct device *dev)
{
    ei_open(dev);
    MOD_INC_USE_COUNT; /* Module stuff ?*/
    return 0;
}

static int
ne_close(struct device *dev)
{
    if (ei_debug > 1)
	printk("%s: Shutting down ethercard.\n", dev->name);
    ei_close(dev);
    MOD_DEC_USE_COUNT; /*Module stuff ?*/
    return 0;
}

/* Hard reset the card.  This used to pause for the same period that a
   8390 reset command required, but that shouldn't be necessary. */
static void
ne_reset_8390(struct device *dev)
{
    unsigned long reset_start_time = jiffies;
    short i;
/*  printk("ne_reset_8390: Called !\n");*/

    if (ei_debug > 1) printk("resetting the 8390 t=%ld...", jiffies);
#ifdef OLDCODE
    /* DON'T change these to inb_p/outb_p or reset will fail on clones. */
    outb(inb(NE_BASE + NE_RESET), NE_BASE + NE_RESET);

    ei_status.txing = 0;
    ei_status.dmaing = 0;

    /* This check _should_not_ be necessary, omit eventually. */
    while ((inb_p(NE_BASE+EN0_ISR) & ENISR_RESET) == 0)
	if (jiffies - reset_start_time > 2*HZ/100) {
	    printk("%s: ne_reset_8390() did not complete.\n", dev->name);
	    break;
	}
    outb_p(ENISR_RESET, NE_BASE + EN0_ISR);	/* Ack intr. */
#endif
    outb(0x21, NE_BASE + NE_CMD);   /*Stop the damnd thing*/
    outb(0xff, NE_BASE + ISR);      /*All interrupts handled !*/
    outb(0x00, NE_BASE + IMR);      /* */
    for (i=0; i<8; i++) outb(0x0,NE_BASE + NE_RESET); /*Not sure if this is ok*/  
    inb(NE_BASE + NE_RESET); /*Reset as they did it in the crynwr stuff*/
    outb(0x21, NE_BASE + NE_CMD);   /*Kick it*/

    ei_status.txing = 0;
    ei_status.dmaing = 0;

    /* This check _should_not_ be necessary, omit eventually. */
    while ((inb_p(NE_BASE+EN0_ISR) & ENISR_RESET) == 0)
	if (jiffies - reset_start_time > 2*HZ/100) {
	    printk("%s: ne_reset_8390() did not complete.\n", dev->name);
	    break;
	}
}

/* Grab the 8390 specific header. Similar to the block_input routine, but
   we don't need to be concerned with ring wrap as the header will be at
   the start of a page, so we optimize accordingly. */

static void
ne_get_8390_hdr(struct device *dev, struct e8390_pkt_hdr *hdr, int ring_page)
{

    int nic_base = dev->base_addr;
  /*printk("ne_get_8390_hdr : called !!\n");*/

    /* This *shouldn't* happen. If it does, it's the last thing you'll see */
    if (ei_status.dmaing) {
	printk("%s: DMAing conflict in ne_get_8390_hdr "
	   "[DMAstat:%d][irqlock:%d][intr:%d].\n",
	   dev->name, ei_status.dmaing, ei_status.irqlock,
	   dev->interrupt);
	return;
    }

    ei_status.dmaing |= 0x01;
    outb_p(E8390_NODMA+E8390_PAGE0+E8390_START, nic_base+ NE_CMD);
    outb_p(sizeof(struct e8390_pkt_hdr), nic_base + EN0_RCNTLO);
    outb_p(0, nic_base + EN0_RCNTHI);
    outb_p(0, nic_base + EN0_RSARLO);		/* On page boundary */
    outb_p(ring_page, nic_base + EN0_RSARHI);
    outb_p(E8390_RREAD+E8390_START, nic_base + NE_CMD);

    /*if (ei_status.word16)*/
	insw(NE_BASE + NE_DATAPORT, hdr, sizeof(struct e8390_pkt_hdr)>>1);
    /*else
	insb(NE_BASE + NE_DATAPORT, hdr, sizeof(struct e8390_pkt_hdr));*/

    outb_p(ENISR_RDC, nic_base + EN0_ISR);	/* Ack intr. */
    ei_status.dmaing &= ~0x01;
}

/* Block input and output, similar to the Crynwr packet driver.  If you
   are porting to a new ethercard, look at the packet driver source for hints.
   The NEx000 doesn't share the on-board packet memory -- you have to put
   the packet out through the "remote DMA" dataport using outb. */

static void
ne_block_input(struct device *dev, int count, struct sk_buff *skb, int ring_offset)
{
#ifdef NE_SANITY_CHECK
    int xfer_count = count;
#endif
    int nic_base = dev->base_addr;
    char *buf = skb->data;

/*printk("ne_block_input : called !\n");*/

    /* This *shouldn't* happen. If it does, it's the last thing you'll see */
    if (ei_status.dmaing) {
	printk("%s: DMAing conflict in ne_block_input "
	   "[DMAstat:%d][irqlock:%d][intr:%d].\n",
	   dev->name, ei_status.dmaing, ei_status.irqlock,
	   dev->interrupt);
	return;
    }
    ei_status.dmaing |= 0x01;
    outb_p(E8390_NODMA+E8390_PAGE0+E8390_START, nic_base+ NE_CMD);
    outb_p(count & 0xff, nic_base + EN0_RCNTLO);
    outb_p(count >> 8, nic_base + EN0_RCNTHI);
    outb_p(ring_offset & 0xff, nic_base + EN0_RSARLO);
    outb_p(ring_offset >> 8, nic_base + EN0_RSARHI);
    outb_p(E8390_RREAD+E8390_START, nic_base + NE_CMD);
/*    if (ei_status.word16) {*/
      insw(NE_BASE + NE_DATAPORT,buf,count>>1);
      if (count & 0x01) {
	buf[count-1] = inb(NE_BASE + NE_DATAPORT);
#ifdef NE_SANITY_CHECK
	xfer_count++;
#endif
      }
/*    } else {
	insb(NE_BASE + NE_DATAPORT, buf, count);
    }*/

#ifdef NE_SANITY_CHECK
    /* This was for the ALPHA version only, but enough people have
       been encountering problems so it is still here.  If you see
       this message you either 1) have a slightly incompatible clone
       or 2) have noise/speed problems with your bus. */
    if (ei_debug > 1) {		/* DMA termination address check... */
	int addr, tries = 20;
	do {
	    /* DON'T check for 'inb_p(EN0_ISR) & ENISR_RDC' here
	       -- it's broken for Rx on some cards! */
	    int high = inb_p(nic_base + EN0_RSARHI);
	    int low = inb_p(nic_base + EN0_RSARLO);
	    addr = (high << 8) + low;
	    if (((ring_offset + xfer_count) & 0xff) == low)
		break;
	} while (--tries > 0);
	if (tries <= 0)
	    printk("%s: RX transfer address mismatch,"
		   "%#4.4x (expected) vs. %#4.4x (actual).\n",
		   dev->name, ring_offset + xfer_count, addr);
    }
#endif
    outb_p(ENISR_RDC, nic_base + EN0_ISR);	/* Ack intr. */
    ei_status.dmaing &= ~0x01;
}

static void
ne_block_output(struct device *dev, int count,
		const unsigned char *buf, const int start_page)
{
    int nic_base = NE_BASE;
    unsigned long dma_start;
#ifdef NE_SANITY_CHECK
    int retries = 0;
#endif
  
  /*printk("ne_block_output : Called !, word16 = %d\n", ei_status.word16);*/
  /*if (ei_status.word16) printk("in WORD mode\n");*/

    /* Round the count up for word writes.  Do we need to do this?
       What effect will an odd byte count have on the 8390?
       I should check someday. */
    if (ei_status.word16 && (count & 0x01))
      count++;

    /* This *shouldn't* happen. If it does, it's the last thing you'll see */
    if (ei_status.dmaing) {
	printk("%s: DMAing conflict in ne_block_output."
	   "[DMAstat:%d][irqlock:%d][intr:%d]\n",
	   dev->name, ei_status.dmaing, ei_status.irqlock,
	   dev->interrupt);
	return;
    }
    ei_status.dmaing |= 0x01;
    /* We should already be in page 0, but to be safe... */
    outb_p(E8390_PAGE0+E8390_START+E8390_NODMA, nic_base + NE_CMD);

#ifdef NE_SANITY_CHECK
 retry:
#endif

#ifdef NE8390_RW_BUGFIX
    /* Handle the read-before-write bug the same way as the
       Crynwr packet driver -- the NatSemi method doesn't work.
       Actually this doesn't always work either, but if you have
       problems with your NEx000 this is better than nothing! */
    outb_p(0x42, nic_base + EN0_RCNTLO);
    outb_p(0x00,   nic_base + EN0_RCNTHI);
    outb_p(0x42, nic_base + EN0_RSARLO);
    outb_p(0x00, nic_base + EN0_RSARHI);
    outb_p(E8390_RREAD+E8390_START, nic_base + NE_CMD);
    /* Make certain that the dummy read has occurred. */
    SLOW_DOWN_IO;
    SLOW_DOWN_IO;
    SLOW_DOWN_IO;
#endif

    outb_p(ENISR_RDC, nic_base + EN0_ISR);

   /* Now the normal output. */
    outb_p(count & 0xff, nic_base + EN0_RCNTLO);
    outb_p(count >> 8,   nic_base + EN0_RCNTHI);
    outb_p(0x00, nic_base + EN0_RSARLO);
    outb_p(start_page, nic_base + EN0_RSARHI);

    outb_p(E8390_RWRITE+E8390_START, nic_base + NE_CMD);
    /*if (ei_status.word16) {*/
	outsw(NE_BASE + NE_DATAPORT, buf, count>>1);
    /*} else {
	outsb(NE_BASE + NE_DATAPORT, buf, count);
    }*/

    dma_start = jiffies;

#ifdef NE_SANITY_CHECK
    /* This was for the ALPHA version only, but enough people have
       been encountering problems so it is still here. */
    if (ei_debug > 1) {		/* DMA termination address check... */
	int addr, tries = 20;
	do {
	    int high = inb_p(nic_base + EN0_RSARHI);
	    int low = inb_p(nic_base + EN0_RSARLO);
	    addr = (high << 8) + low;
	    if ((start_page << 8) + count == addr)
		break;
	} while (--tries > 0);
	if (tries <= 0) {
	    printk("%s: Tx packet transfer address mismatch,"
		   "%#4.4x (expected) vs. %#4.4x (actual).\n",
		   dev->name, (start_page << 8) + count, addr);
	    if (retries++ == 0)
		goto retry;
	}
    }
#endif

    while ((inb_p(nic_base + EN0_ISR) & ENISR_RDC) == 0)
	if (jiffies - dma_start > 2*HZ/100) {		/* 20ms */
		printk("%s: timeout waiting for Tx RDC.\n", dev->name);
		ne_reset_8390(dev);
		NS8390_init(dev,1);
		break;
	}

    outb_p(ENISR_RDC, nic_base + EN0_ISR);	/* Ack intr. */
    ei_status.dmaing &= ~0x01;
/*printk("Output done\n");*/
    return;
}


#ifdef MODULE_NOT_YET_SUPORTED /*not yet suported :)*/
#define MAX_NE_CARDS	4	/* Max number of NE cards per module */
#define NAMELEN		8	/* # of chars for storing dev->name */
static char namelist[NAMELEN * MAX_NE_CARDS] = { 0, };
static struct device dev_ne[MAX_NE_CARDS] = {
	{
		NULL,		/* assign a chunk of namelist[] below */
		0, 0, 0, 0,
		0, 0,
		0, 0, 0, NULL, NULL
	},
};

static int io[MAX_NE_CARDS] = { 0, };
static int irq[MAX_NE_CARDS]  = { 0, };

/* This is set up so that no autoprobe takes place. We can't guarantee
that the ne2k probe is the last 8390 based probe to take place (as it
is at boot) and so the probe will get confused by any other 8390 cards.
ISA device autoprobes on a running machine are not recommended anyway. */

int
init_module(void)
{
	int this_dev, found = 0;

	for (this_dev = 0; this_dev < MAX_NE_CARDS; this_dev++) {
		struct device *dev = &dev_ne[this_dev];
		dev->name = namelist+(NAMELEN*this_dev);
		dev->irq = irq[this_dev];
		dev->base_addr = io[this_dev];
		dev->init = ne_probe;
		if (io[this_dev] == 0)  {
			if (this_dev != 0) break; /* only complain once */
			printk(KERN_NOTICE "ne2-mca.c: Module autoprobing not allowed. Append \"io=0xNNN\" value(s).\n");
			return -EPERM;
		}
		if (register_netdev(dev) != 0) {
			printk(KERN_WARNING "nei2.c: No NE*000 card found (i/o = 0x%x).\n", io[this_dev]);
			if (found != 0) return 0;	/* Got at least one. */
			return -ENXIO;
		}
		found++;
	}

	return 0;
}

void
cleanup_module(void)
{
	int this_dev;

	for (this_dev = 0; this_dev < MAX_NE_CARDS; this_dev++) {
		struct device *dev = &dev_ne[this_dev];
		if (dev->priv != NULL) {
			kfree(dev->priv);
			dev->priv = NULL;
			free_irq(dev->irq);
			irq2dev_map[dev->irq] = NULL;
			release_region(dev->base_addr, NE_IO_EXTENT);
			unregister_netdev(dev);
		}
	}
}
#endif /* MODULE */

/*
 * Local variables:
 *  compile-command: "gcc -DKERNEL -Wall -O6 -fomit-frame-pointer -I/usr/src/linux/net/tcp -c ne2-mca.c"
 */
